@@ -5,11 +5,12 @@
 
 package Future;
 
+use v5.10;
 use strict;
 use warnings;
 no warnings 'recursion'; # Disable the "deep recursion" warning
 
-our $VERSION = '0.43';
+our $VERSION = '0.47';
 
 use Carp qw(); # don't import croak
 use Scalar::Util qw( weaken blessed reftype );
@@ -25,6 +26,8 @@ our @CARP_NOT = qw( Future::Utils );
 
 use constant DEBUG => !!$ENV{PERL_FUTURE_DEBUG};
 
+use constant STRICT => !!$ENV{PERL_FUTURE_STRICT};
+
 our $TIMES = DEBUG || $ENV{PERL_FUTURE_TIMES};
 
 =head1 NAME
@@ -132,8 +135,8 @@ instance.
       return $self;
    }
 
-If an instance overrides the L</block_until_ready> method, this will be called
-by C<get> and C<failure> if the instance is still pending.
+If an instance overrides the L</await> method, this will be called by C<get>
+and C<failure> if the instance is still pending.
 
 In most cases this should allow future-returning modules to be used as if they
 were blocking call/return-style modules, by simply appending a C<get> call to
@@ -224,6 +227,8 @@ use constant {
 
    CB_SEQ_IMDONE => 1<<7, # $code is in fact immediate ->done result
    CB_SEQ_IMFAIL => 1<<8, # $code is in fact immediate ->fail result
+
+   CB_SEQ_STRICT => 1<<9, # Complain if $code didn't return a Future
 };
 
 use constant CB_ALWAYS => CB_DONE|CB_FAIL|CB_CANCEL;
@@ -270,7 +275,7 @@ sub new
    }, ( ref $proto || $proto );
 }
 
-*AWAIT_CLONE = sub { shift->new };  # We need to respect subclassing
+*AWAIT_CLONE = sub { shift->new };
 
 my $GLOBAL_END;
 END { $GLOBAL_END = 1; }
@@ -395,8 +400,8 @@ sub _mark_ready
    my $fail      = defined $self->{failure};
    my $done      = !$fail && !$cancelled;
 
-   my @result  = $done ? $self->get :
-                 $fail ? $self->failure :
+   my @result  = $done ? @{ $self->{result} } :
+                 $fail ? @{ $self->{failure} } :
                          ();
 
    foreach my $cb ( @$callbacks ) {
@@ -449,8 +454,12 @@ sub _mark_ready
             }
 
             unless( blessed $f2 and $f2->isa( "Future" ) ) {
-               $fseq->fail( "Expected " . CvNAME_FILE_LINE($code) . " to return a Future" );
-               next;
+               # Upgrade a non-Future result, or complain in strict mode
+               if( $flags & CB_SEQ_STRICT ) {
+                  $fseq->fail( "Expected " . CvNAME_FILE_LINE($code) . " to return a Future" );
+                  next;
+               }
+               $f2 = Future->done( $f2 );
             }
 
             $fseq->on_cancel( $f2 );
@@ -509,7 +518,7 @@ sub is_ready
    return $self->{ready};
 }
 
-*AWAIT_IS_READY = \&is_ready;
+*AWAIT_IS_READY = sub { shift->is_ready };
 
 =head2 is_done
 
@@ -557,7 +566,7 @@ sub is_cancelled
    return $self->{cancelled};
 }
 
-*AWAIT_IS_CANCELLED = \&is_cancelled;
+*AWAIT_IS_CANCELLED = sub { shift->is_cancelled };
 
 =head2 state
 
@@ -599,6 +608,9 @@ Cannot be called on a convergent future.
 If the future is already cancelled, this request is ignored. If the future is
 already complete with a result or a failure, an exception is thrown.
 
+I<Since version 0.45:> this method is also available under the name
+C<resolve>.
+
 =cut
 
 sub done
@@ -622,9 +634,11 @@ sub done
    return $self;
 }
 
+*resolve = sub { shift->done( @_ ) };
+
 # TODO: For efficiency we can implement better versions of these as individual
 #  methods know which case is being invoked
-*AWAIT_NEW_DONE = *AWAIT_DONE = \&done;
+*AWAIT_NEW_DONE = *AWAIT_DONE = sub { shift->done( @_ ) };
 
 =head2 fail
 
@@ -649,6 +663,8 @@ additional details to be transparently preserved by such code as
       return Future->fail($@);
    }
 
+I<Since version 0.45:> this method is also available under the name C<reject>.
+
 =cut
 
 sub fail
@@ -680,9 +696,11 @@ sub fail
    return $self;
 }
 
+*reject = sub { shift->fail( @_ ) };
+
 # TODO: For efficiency we can implement better versions of these as individual
 #  methods know which case is being invoked
-*AWAIT_NEW_FAIL = *AWAIT_FAIL = \&fail;
+*AWAIT_NEW_FAIL = *AWAIT_FAIL = sub { shift->fail( @_ ) };
 
 =head2 die
 
@@ -764,6 +782,9 @@ sub AWAIT_ON_CANCEL
    weaken( $r->[1] );
 }
 
+# New name for it
+*AWAIT_CHAIN_CANCEL = \&AWAIT_ON_CANCEL;
+
 sub _revoke_on_cancel
 {
    my $self = shift;
@@ -828,8 +849,8 @@ sub on_ready
 
       $self->{reported} = 1 if $fail;
 
-      $is_future ? ( $done ? $code->done( $self->get ) :
-                     $fail ? $code->fail( $self->failure ) :
+      $is_future ? ( $done ? $code->done( @{ $self->{result} } ) :
+                     $fail ? $code->fail( @{ $self->{failure} } ) :
                              $code->cancel )
                  : $code->( $self );
    }
@@ -848,11 +869,13 @@ sub AWAIT_ON_READY
    push @{ $self->{callbacks} }, [ CB_ALWAYS|CB_SELF, $self->wrap_cb( on_ready => $code ) ];
 }
 
-=head2 get
+=head2 result
 
-   @result = $future->get
+   @result = $future->result
 
-   $result = $future->get
+   $result = $future->result
+
+I<Since version 0.44.>
 
 If the future is ready and completed successfully, returns the list of
 results that had earlier been given to the C<done> method on a leaf future,
@@ -864,17 +887,15 @@ failure that was given to the C<fail> method. If additional details were given
 to the C<fail> method, an exception object is constructed to wrap them of type
 L<Future::Exception>.
 
-If the future was cancelled an exception is thrown.
-
-If it is not yet ready then L</block_until_ready> is invoked to wait for a
-ready state.
+If the future was cancelled or is not yet ready an exception is thrown.
 
 =cut
 
-sub get
+sub result
 {
    my $self = shift;
-   $self->block_until_ready unless $self->{ready};
+   $self->{ready} or
+      Carp::croak( "${\$self->__selfstr} is not yet ready" );
    if( my $failure = $self->{failure} ) {
       $self->{reported} = 1;
       my $exception = $failure->[0];
@@ -886,15 +907,34 @@ sub get
    return @{ $self->{result} };
 }
 
-# TODO: For efficiency we can implement a better version of this because it
-#   is only invoked on ready futures
-*AWAIT_GET = \&get;
+*AWAIT_RESULT = *AWAIT_GET = sub { shift->result };
 
-=head2 block_until_ready
+=head2 get
 
-   $f = $f->block_until_ready
+   @result = $future->get
 
-I<Since version 0.40.>
+   $result = $future->get
+
+If the future is ready, returns the result or throws the failure exception as
+per L</result>.
+
+If it is not yet ready then L</await> is invoked to wait for a ready state, and
+the result returned as above.
+
+=cut
+
+sub get
+{
+   my $self = shift;
+   $self->await unless $self->{ready};
+   return $self->result;
+}
+
+=head2 await
+
+   $f = $f->await
+
+I<Since version 0.44.>
 
 Blocks until the future instance is no longer pending.
 
@@ -906,30 +946,36 @@ method is useful in cases where the exception-throwing part of C<get> is not
 required, perhaps because other code will be testing the result using
 L</is_done> or similar.
 
-   if( $f->block_until_ready->is_done ) {
+   if( $f->await->is_done ) {
       ...
    }
 
-This method is intended for subclasses to override, but a default
-implementation for back-compatibility purposes is provided which calls the
-C<await> method. If the future is not yet ready, attempts to wait for an
-eventual result by using the underlying C<await> method, which subclasses
-should provide. The default implementation will throw an exception if called
-on a still-pending instance that does not provide an C<await> method.
+This method is intended for subclasses to override. The default implementation
+will throw an exception if called on a still-pending instance.
 
 =cut
 
 sub await
 {
    my $self = shift;
+   return $self if $self->{ready};
    Carp::croak "$self is not yet complete and does not provide ->await";
 }
 
+=head2 block_until_ready
+
+   $f = $f->block_until_ready
+
+I<Since version 0.40.>
+
+Now a synonym for L</await>. New code should invoke C<await> directly.
+
+=cut
+
 sub block_until_ready
 {
    my $self = shift;
-   until( $self->{ready} ) { $self->await }
-   return $self;
+   return $self->await;
 }
 
 =head2 unwrap
@@ -942,7 +988,7 @@ If given a single argument which is a C<Future> reference, this method will
 call C<get> on it and return the result. Otherwise, it returns the list of
 values directly in list context, or the first value in scalar. Since it
 involves an implicit blocking wait, this method can only be used on immediate
-futures or subclasses that implement L</block_until_ready>.
+futures or subclasses that implement L</await>.
 
 This will ensure that an outgoing argument is definitely not a C<Future>, and
 may be useful in such cases as adapting synchronous code to fit asynchronous
@@ -996,8 +1042,8 @@ sub on_done
    if( $self->{ready} ) {
       return $self if $self->{failure} or $self->{cancelled};
 
-      $is_future ? $code->done( $self->get ) 
-                 : $code->( $self->get );
+      $is_future ? $code->done( @{ $self->{result} } ) 
+                 : $code->( @{ $self->{result} } );
    }
    else {
       push @{ $self->{callbacks} }, [ CB_DONE|CB_RESULT, $self->wrap_cb( on_done => $code ) ];
@@ -1015,8 +1061,7 @@ sub on_done
 If the future is ready, returns the exception passed to the C<fail> method or
 C<undef> if the future completed successfully via the C<done> method.
 
-If it is not yet ready then L</block_until_ready> is invoked to wait for a
-ready state.
+If it is not yet ready then L</await> is invoked to wait for a ready state.
 
 If called in list context, will additionally yield the category name and list
 of the details provided to the C<fail> method.
@@ -1028,7 +1073,7 @@ statement:
       ...
    }
    else {
-      my @result = $future->get;
+      my @result = $future->result;
       ...
    }
 
@@ -1037,7 +1082,7 @@ statement:
 sub failure
 {
    my $self = shift;
-   $self->block_until_ready unless $self->{ready};
+   $self->await unless $self->{ready};
    return unless $self->{failure};
    $self->{reported} = 1;
    return $self->{failure}->[0] if !wantarray;
@@ -1083,8 +1128,8 @@ sub on_fail
       return $self if not $self->{failure};
       $self->{reported} = 1;
 
-      $is_future ? $code->fail( $self->failure )
-                 : $code->( $self->failure );
+      $is_future ? $code->fail( @{ $self->{failure} } )
+                 : $code->( @{ $self->{failure} } );
    }
    else {
       push @{ $self->{callbacks} }, [ CB_FAIL|CB_RESULT, $self->wrap_cb( on_fail => $code ) ];
@@ -1135,13 +1180,28 @@ code depending on the success or failure of the first, or may run it
 regardless. The returned sequence future represents the entire combination of
 activity.
 
-In some cases the code should return a future; in some it should return an
-immediate result. If a future is returned, the combined future will then wait
-for the result of this second one. If the combinined future is cancelled, it
-will cancel either the first future or the second, depending whether the first
-had completed. If the code block throws an exception instead of returning a
-value, the sequence future will fail with that exception as its message and no
-further values.
+The invoked code could return a future, or a result directly.
+
+I<Since version 0.45:> if a non-future result is returned it will be wrapped
+in a new immediate Future instance. This behaviour can be disabled by setting
+the C<PERL_FUTURE_STRICT> environment variable to a true value at compiletime:
+
+   $ PERL_FUTURE_STRICT=1 perl ...
+
+The combined future will then wait for the result of this second one. If the
+combinined future is cancelled, it will cancel either the first future or the
+second, depending whether the first had completed. If the code block throws an
+exception instead of returning a value, the sequence future will fail with
+that exception as its message and no further values.
+
+Note that since the code is invoked in scalar context, you cannot directly
+return a list of values this way. Any list-valued results must be done by
+returning a C<Future> instance.
+
+   sub {
+      ...
+      return Future->done( @results );
+   }
 
 As it is always a mistake to call these sequencing methods in void context and lose the
 reference to the returned future (because exception/error handling would be
@@ -1154,6 +1214,8 @@ sub _sequence
    my $f1 = shift;
    my ( $code, $flags ) = @_;
 
+   $flags |= CB_SEQ_STRICT if STRICT;
+
    # For later, we might want to know where we were called from
    my $func = (caller 1)[3];
    $func =~ s/^.*:://;
@@ -1167,8 +1229,8 @@ sub _sequence
 
    if( $f1->is_ready ) {
       # Take a shortcut
-      return $f1 if $f1->is_done and not( $flags & CB_SEQ_ONDONE ) or
-                    $f1->failure and not( $flags & CB_SEQ_ONFAIL );
+      return $f1 if $f1->is_done   and not( $flags & CB_SEQ_ONDONE ) or
+                    $f1->{failure} and not( $flags & CB_SEQ_ONFAIL );
 
       if( $flags & CB_SEQ_IMDONE ) {
          return Future->done( @$code );
@@ -1179,8 +1241,8 @@ sub _sequence
 
       my @args = (
          ( $flags & CB_SELF ? $f1 : () ),
-         ( $flags & CB_RESULT ? $f1->is_done ? $f1->get :
-                                $f1->failure ? $f1->failure :
+         ( $flags & CB_RESULT ? $f1->is_done   ? @{ $f1->{result} } :
+                                $f1->{failure} ? @{ $f1->{failure} } :
                                                () : () ),
       );
 
@@ -1190,7 +1252,11 @@ sub _sequence
       }
 
       unless( blessed $fseq and $fseq->isa( "Future" ) ) {
-         return Future->fail( "Expected " . CvNAME_FILE_LINE($code) . " to return a Future" );
+         # Upgrade a non-Future result, or complain in strict mode
+         $flags & CB_SEQ_STRICT and
+            return Future->fail( "Expected " . CvNAME_FILE_LINE($code) . " to return a Future" );
+
+         $fseq = $f1->new->done( $fseq );
       }
 
       return $fseq;
@@ -1246,8 +1312,8 @@ with the same result and the code will not be invoked.
 The C<then> method can also be passed the C<$fail_code> block as well, giving
 a combination of C<then> and C<else> behaviour.
 
-This operation is designed to be compatible with the semantics of other future
-systems, such as Javascript's Q or Promises/A libraries.
+This operation is similar to those provided by other future systems, such as
+Javascript's Q or Promises/A libraries.
 
 =cut
 
@@ -1273,15 +1339,15 @@ my $make_donecatchfail_sub = sub {
 
       if( !$self->{failure} ) {
          return $self unless $done_code;
-         return $done_code->( @maybe_self, $self->get );
+         return $done_code->( @maybe_self, @{ $self->{result} } );
       }
       else {
          my $name = $self->{failure}[1];
          if( defined $name and $catch_handlers{$name} ) {
-            return $catch_handlers{$name}->( @maybe_self, $self->failure );
+            return $catch_handlers{$name}->( @maybe_self, @{ $self->{failure} } );
          }
          return $self unless $fail_code;
-         return $fail_code->( @maybe_self, $self->failure );
+         return $fail_code->( @maybe_self, @{ $self->{failure} } );
       }
    };
 };
@@ -1415,12 +1481,12 @@ sub transform
       my $self = shift;
       if( !$self->{failure} ) {
          return $self unless $xfrm_done;
-         my @result = $xfrm_done->( $self->get );
+         my @result = $xfrm_done->( @{ $self->{result} } );
          return $self->new->done( @result );
       }
       else {
          return $self unless $xfrm_fail;
-         my @failure = $xfrm_fail->( $self->failure );
+         my @failure = $xfrm_fail->( @{ $self->{failure} } );
          return $self->new->fail( @failure );
       }
    }, CB_SEQ_ONDONE|CB_SEQ_ONFAIL|CB_SELF );
@@ -1601,6 +1667,10 @@ original. This may be useful if the original future represents an operation
 that is being shared among multiple sequences; cancelling one should not
 prevent the others from running too.
 
+Note that this only prevents cancel propagating from C<$future> to C<$f1>; if
+the original C<$f1> instance is cancelled then the returned C<$future> will
+have to be cancelled too.
+
 =cut
 
 sub without_cancel
@@ -1610,11 +1680,14 @@ sub without_cancel
 
    $self->on_ready( sub {
       my $self = shift;
-      if( $self->failure ) {
-         $new->fail( $self->failure );
+      if( $self->{cancelled} ) {
+         $new->cancel;
+      }
+      elsif( $self->{failure} ) {
+         $new->fail( @{ $self->{failure} } );
       }
       else {
-         $new->done( $self->get );
+         $new->done( @{ $self->{result} } );
       }
    });
 
@@ -1797,10 +1870,10 @@ sub wait_any
       }
 
       if( $immediate_ready->{failure} ) {
-         $self->{failure} = [ $immediate_ready->failure ];
+         $self->{failure} = [ @{ $immediate_ready->{failure} } ];
       }
       else {
-         $self->{result} = [ $immediate_ready->get ];
+         $self->{result} = [ @{ $immediate_ready->{result} } ];
       }
       $self->_mark_ready( "wait_any" );
       return $self;
@@ -1819,10 +1892,10 @@ sub wait_any
          $self->{failure} = [ "All component futures were cancelled" ];
       }
       elsif( $_[0]->{failure} ) {
-         $self->{failure} = [ $_[0]->failure ];
+         $self->{failure} = [ @{ $_[0]->{failure} } ];
       }
       else {
-         $self->{result}  = [ $_[0]->get ];
+         $self->{result}  = [ @{ $_[0]->{result} } ];
       }
 
       foreach my $sub ( @subs ) {
@@ -1888,7 +1961,7 @@ sub needs_all
          $sub->{ready} or $sub->cancel;
       }
 
-      $self->{failure} = [ $immediate_fail->failure ];
+      $self->{failure} = [ @{ $immediate_fail->{failure} } ];
       $self->_mark_ready( "needs_all" );
       return $self;
    }
@@ -1898,7 +1971,7 @@ sub needs_all
 
    # Look for immediate done
    if( !$pending ) {
-      $self->{result} = [ map { $_->get } @subs ];
+      $self->{result} = [ map { @{ $_->{result} } } @subs ];
       $self->_mark_ready( "needs_all" );
       return $self;
    }
@@ -1915,8 +1988,8 @@ sub needs_all
          }
          $self->_mark_ready( "needs_all" );
       }
-      elsif( my @failure = $_[0]->failure ) {
-         $self->{failure} = \@failure;
+      elsif( $_[0]->{failure} ) {
+         $self->{failure} = [ @{ $_[0]->{failure} } ];
          foreach my $sub ( @subs ) {
             $sub->cancel if !$sub->{ready};
          }
@@ -1926,7 +1999,7 @@ sub needs_all
          $pending--;
          $pending and return;
 
-         $self->{result} = [ map { $_->get } @subs ];
+         $self->{result} = [ map { @{ $_->{result} } } @subs ];
          $self->_mark_ready( "needs_all" );
       }
    };
@@ -1992,7 +2065,7 @@ sub needs_any
          $sub->{ready} ? $sub->{reported} = 1 : $sub->cancel;
       }
 
-      $self->{result} = [ $immediate_done->get ];
+      $self->{result} = [ @{ $immediate_done->{result} } ];
       $self->_mark_ready( "needs_any" );
       return $self;
    }
@@ -2022,14 +2095,14 @@ sub needs_any
          $self->{failure} = [ "All component futures were cancelled" ];
          $self->_mark_ready( "needs_any" );
       }
-      elsif( my @failure = $_[0]->failure ) {
+      elsif( $_[0]->{failure} ) {
          $pending and return;
 
-         $self->{failure} = \@failure;
+         $self->{failure} = [ @{ $_[0]->{failure} } ];
          $self->_mark_ready( "needs_any" );
       }
       else {
-         $self->{result} = [ $_[0]->get ];
+         $self->{result} = [ @{ $_[0]->{result} } ];
          foreach my $sub ( @subs ) {
             $sub->cancel if !$sub->{ready};
          }
@@ -2316,7 +2389,7 @@ method, and obtain the result using C<get>.
 
    $f->on_ready( sub {
       my $f = shift;
-      say "The operation returned: ", $f->get;
+      say "The operation returned: ", $f->result;
    } );
 
 =head2 Indicating Success or Failure
@@ -2330,7 +2403,7 @@ failure.
    $f->on_ready( sub {
       my $f = shift;
       if( not my $e = $f->failure ) {
-         say "The operation succeeded with: ", $f->get;
+         say "The operation succeeded with: ", $f->result;
       }
       else {
          say "The operation failed with: ", $e;
@@ -2349,7 +2422,7 @@ called I<Exception Hoisting>).
    $f->on_ready( sub {
       my $f = shift;
       try {
-         say "The operation succeeded with: ", $f->get;
+         say "The operation succeeded with: ", $f->result;
       }
       catch {
          say "The operation failed with: ", $_;
@@ -2425,8 +2498,8 @@ for multiple concurrent operations to finish.
 
    $f->on_ready( sub {
       say "Operations are ready:";
-      say "  foo: ", $f1->get;
-      say "  bar: ", $f2->get;
+      say "  foo: ", $f1->result;
+      say "  bar: ", $f2->result;
    } );
 
 This provides an ability somewhat similar to C<CPS::kpar()> or
